// RouteHelper.java

import java.util.*;
import java.io.*;
import java.awt.*;

import transport.*;
import transport.protocol.*;
import util.*;

/**
 * RouteHelper is a server that runs independently on port given by
 * ROUTEHELPER_PORT. The client sends a <code>PosChangedMessage</code>
 * indicating change in position. Based on thr<code>route</code> and current
 * position of the client, RouteHelper replies with a <code>
 * RouteHelperMessage</code> that tells the client whether it is on or off
 * track.
 *
 * @author Amit G.
 */
public class RouteHelper implements SocketReceiverListener {

  public static final int ROUTEHELPER_PORT  = 54021;
  public static final int INTERVAL = 40; // increase for higher accuracy

  /** 
   * Smaller the value of THRESH, closer the user must be to the route for the
   * server to return "on-track". A value of zero indicates that the user
   * must be exactly on the route.
   */
  public static final int THRESH = 20;

  public RouteHelper(int port) throws IOException {
    SocketReceiver s = new SocketReceiver(this, port);
    s.start();
    System.out.println("RouteHelper started on port " + port);
  }
  
  public void incomingMessage(Message m) {
    if (m instanceof PosChangedMessage) {
      PosChangedMessage pmsg = (PosChangedMessage)m;
      if (pmsg.getRouteName() == null || pmsg.getRouteName().equals("")) 
        return;

      // create the route object
      // (This is probably not a good way of doing it since Route has
      // to do all the XML processing in this constructor. Perhaps a better
      // approach would be to use a RouteFactory that maintains a certain
      // number of instantiated routes and returns them on demand)
      Route route = new Route(new File(pmsg.getRouteName()));

      // validate the current position and send the reply
      RouteHelperMessage rmsg = new RouteHelperMessage(pmsg.getFrom(), 
                                                       ROUTEHELPER_PORT,
                                                       check(route, 
                                                             pmsg.getPoint()));

      SocketSender ss = new SocketSender(rmsg);
      ss.start();
    }
  }

  public static void main(String args[]) throws Exception {
    new RouteHelper(ROUTEHELPER_PORT);
  }

  private boolean check(Route route, Point currentPoint) {

    /*
    ** algo:
    ** Divide the path between every two consecutive points in the route
    ** into INTERVAL parts. Check if the current position is close to any of
    ** these points. It is close if the distance is less than THRESH
    */
    Vector finePoints = new Vector();
    for (int i = 0; i < route.getPoints().size() - 1; i++) {
      finePoints.addAll(splitPath((Point)route.getPoints().elementAt(i), 
                                  (Point)route.getPoints().elementAt(i + 1)));
    }

    boolean lost = true;
    for (int i = 0; i < finePoints.size(); i++) {
      int len = length(currentPoint, (Point)finePoints.elementAt(i));
      if (len <= THRESH) {
        lost = false;
        break;
      }
    }
    return lost;
  }

  private Vector splitPath(Point p1, Point p2) {
    Vector splits = new Vector();
    
    int partsCount = length(p1, p2) / INTERVAL;
    if (partsCount != 0) { //if  length is not smaller than the interval
      int xInc = (int)(p2.getX() - p1.getX()) / partsCount;//xIncrement
      int yInc = (int)(p2.getY() - p1.getY()) / partsCount;//yIncrement
      
      splits.add(0, new Point(p1));
      for (int i = 1; i < partsCount; i++) {
        p1.translate(xInc, yInc);
        splits.add(i, new Point(p1));
      }
    }
    else {
      splits.add(new Point(p1));
      splits.add(new Point(p2));
      return splits;
    }
    return splits;
  }

  // returns length between two points using pythagoras
  private int length(Point p1, Point p2) {
    double lenX = p1.getX() - p2.getX();
    double lenY = p1.getY() - p2.getY();
    return (int)Math.sqrt(lenX*lenX + lenY*lenY);
  }

} // class RouteHelper
