/**<pre>
 * Title: GPS Simulator
 * Copyright: Copyright (c) 2002
 * Company:   University of Oregon Computer Science Dept.
 *
 * @author Jason Prideaux
 * @version 1.0
 * </pre>
 */

package gui;

import controller.*;
import agent.*;
import util.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.lang.Math;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;



/* ===================================================================== */
/** This class is an extension of a JPanel.  It's backgound is that of
 *  scale map corresponding to real world GPS coordinates.  Agents are drawn
 *  on the JPanel, and can be dragged around the map. also wirelss ranges
 *  are drawn on each Agent if requested by the iSIM user.
 *
 */
public class ChartComp extends JPanel implements MouseListener, MouseMotionListener{


	/** The controller of the simulation. */
	private Controller controller;

	/** The chart data that contains map display and gps information. */
  	private ChartData chartdata;

	/** True if when user clicks to move an agent on map. */
	private boolean clicked = false;

	/** True if this JComponent is to display wireless ranges. */
	private boolean ranges;

	/** Holds the map image. */
	private Image img;

	/** The point of key/center agent for drawing. */
	private Point pos;

	/** Offset position on an agent when mouse-dragging. */
	private int x_offset = 0;

	/** Offset position on an agent when mouse-dragging. */
	private int y_offset = 0;

	/** A time for keeping track of double-clicking. */
	private Date doubleclick = new Date();

	/** A Dialog Window for display User information. */
	private PropDialog propDialog;

    /** Listeners for this object */
    private Vector listeners = new Vector();

    /* ===================================================================== */
    /** The constructor sets up mouse listeners and sets up its own variables
     *  with the given ones.
     *
     *  @param  controller   The controller of simulation.
     *  @param  ranges   True if displaying wireless ranges.
     *
     */
  	public ChartComp(Controller controller, boolean ranges){

		this.ranges = ranges;
		this.controller=controller;

		// Sign up listeners
		this.addMouseListener(this);
		this.addMouseMotionListener(this);

  	} //Constructor





    /* ===================================================================== */
    /** This method returns our preferred size which should be the size of the
     *  map image that this object displays.
     *
     *  @return  Dimension   The size of this JPanel with map image.
     *
     */
  	synchronized public Dimension getPreferredSize(){

    	if (img!=null){

			// Wait until the image is loaded so we know the size
			int i=0;
			while ((i<30) && (img.getWidth(this)<0)){

				try {this.wait(100);} catch (InterruptedException e) {;}
				i++;
			}


			return new Dimension(img.getWidth(this), img.getHeight(this));

    	}else{

      		return new Dimension(100,100);
    	}

  	} //method: getPreferredSize



    /* ===================================================================== */
    /** This method is used whenever this JPanel gets painted.  The method
     *  draws all agents onto the map, and draws their wireless ranges if
     *  wanted
     *
     *  @param  g   The Graphics object for drawing images and circles.
     *  @param  users   The user agents to draw.
     *  @param  points   The access point agents to draw.
     *
     */
  	synchronized void draw(Graphics g, Vector users, Vector points){

		// Make sure there is an agent in the sim
    		if( pos!=null ){


			Point t1;
			Point t2;
			Toolkit tool = Toolkit.getDefaultToolkit();   //gotta love tool
			Image img2;

			// Draw users
			for(int i=0; i<users.size(); i++ ){
				pos = ((User)users.elementAt(i)).getLocation();

				t1 = ((User)users.elementAt(i)).getLocation();
				t2 = controller.getCurrentAgent().getLocation();

				img2 = getToolkit().getImage("files/agents/"+((User)users.elementAt(i)).getName()+"_"+((int)(((User)users.elementAt(i)).getPosition().getHeading()))+".gif");
				g.drawImage(img2, pos.x, pos.y, this);

				// Draw wireless range circles for the user.
				if(ranges){
					int circles = (int)((((User)users.elementAt(i)).getRange())/10);
					int ind = 0;
					for(int j=10; j<=((User)users.elementAt(i)).getRange(); j+=10){
						Color c = Color.blue;
						g.setColor(new Color(c.getRed(), c.getGreen(), c.getBlue(), 255-ind*(240/circles) ));
						g.drawOval(pos.x-j+15,pos.y-j+15,j*2,j*2);
						ind++;
					}
				}

			}//for


			// Draw access points
			for(int i=0; i< points.size(); i++ ){
				pos = ((Agent)points.elementAt(i)).getLocation();

				t1 = ((Agent)points.elementAt(i)).getLocation();
				t2 = controller.getCurrentAgent().getLocation();

				img2 = getToolkit().getImage("files/agents/accesspoint.gif");
				g.drawImage(img2, pos.x, pos.y, this);

				// Draw wireless range circles for the access point.
				if(ranges){
					int circles = (int)((((Agent)points.elementAt(i)).getRange())/10);
					int ind = 0;
					for(int j=10; j<=((Agent)points.elementAt(i)).getRange(); j+=10){
						Color c = Color.blue;
						g.setColor(new Color(c.getRed(), c.getGreen(), c.getBlue(), 255-ind*(240/circles) ));
						g.drawOval(pos.x-j+8,pos.y-j+6,j*2,j*2);
						ind++;
					}
				}


			} //for

    	} //if

  	} //method: draw





    /* ===================================================================== */
    /** This method is used to make the JPanel re draw itself, usually because
     *  some information or location has caused an Agent to change or created
     *  a new Agent in the Simulator.
     *
     *  @param  pos   The center point, usually point of current agent.
     *  @param  users   The user agents to draw.
     *  @param  points   The access point agents to draw.
     *
     */
  	synchronized public void updatePos(Point pos, Vector users, Vector points){

		// Get grphics to draw
		Graphics g=getGraphics();

		// Make old drawing
		draw(g, users, points);

		// Set the point, it may be null, or real
		this.pos=pos;

		// Now redraw.
		draw(g, users, points);

		g.dispose();

  	} //method: updatePos





    /* ===================================================================== */
    /** This method is used in case of some time out event occurs to re draw
     *  this JPanel.
     *
     */
  	synchronized public void timeOut(){

    	Graphics g=getGraphics();
    	draw(g, controller.getUsers(), controller.getPoints());

    	draw(g, controller.getUsers(), controller.getPoints());
    	g.dispose();

  	} //method: timeOut



    /* ===================================================================== */
    /** This method is used by JPanels for updating its display, it is safe
     *  to just call paint in this JPanel.
     *
     *  @param  g   The Graphics object for drawing images and circles.
     *
     */
  	public void update(Graphics g){
    	paint(g);

  	} //method: update




    /* ===================================================================== */
    /** This method will get the Graphics object for painting this JPanel, and
     *  then call on paint(Graphics g).
     *
     *  @param  g   The Graphics object for drawing images and circles.
     *
     */
  	synchronized public void paint(){
    	Graphics g=getGraphics();
    	paint(g);
    	g.dispose();

  	} //method: paint



    /* ===================================================================== */
    /** This method is used to paint this JPanel, it will first make sure the
     *  map image is set, and then will call the draw() method to paint all
     *  agents and wireless ranges.
     *
     *  @param  g   The Graphics object for drawing images and circles.
     *
     */
  	synchronized public void paint(Graphics g){

		if (img!=null) {
			g.drawImage(img, 0, 0, this);
		}


		g.setPaintMode();
		g.setColor(Color.blue);

		draw(g,controller.getUsers(), controller.getPoints());
        
        // notify the listeners
        fireGraphicsPaintedEvent(g);

        g.dispose();
	} //method: paint






    /* ===================================================================== */
    /** This method checks for two kinds of mouse clicks.  If first checks for
     *  a right mouse click which is for rotating a user.(changing user's
     *  facing direction)
     *  This mehtod also checks for mouse left button double clicks.  Based on
     *  the location of click and the timing, it will lauch a PropDialog for
     *  the User that was double clicked on.
     *
     *  @param  e   The incoming MouseEvent containing button and point info.
     *
     */
	public void mouseClicked(MouseEvent e){

		int btn = e.getButton();
		Point p = e.getPoint();

		// Check for right click.
		if( btn == 3 ){

			// If we currently dragging the Agent, then rotate if a User.
			if(clicked){

				if( controller.getCurrentAgent() instanceof User ){

					double new_heading = controller.getCurrentAgent().getPosition().getHeading() + 45;

					if( new_heading == 360 )
						new_heading = 0;

					controller.getCurrentAgent().getPosition().setHeading( new_heading );

					this.paint();

				}
			}

		// Check for left click.
		}else if( btn == 1 ){

			double d = 100000;
			double temp;

			// Make sure we clicked on a User or AccessPoint.
			for(int i=0; i<controller.getUsers().size(); i++){

				temp = diff(p, ((User)controller.getUsers().elementAt(i)).getLocation());

				if( temp < d ){
					d = temp;
					controller.setCurrentAgent( ((User)controller.getUsers().elementAt(i)) );
				}

			}
			for(int i=0; i<controller.getPoints().size(); i++){

				temp = diff(p, ((Agent)controller.getPoints().elementAt(i)).getLocation());

				if( temp < d ){
					d = temp;
					controller.setCurrentAgent( ((Agent)controller.getPoints().elementAt(i)) );
				}

			}

			// Now check to see if closest Agent is User, and we are clicking on User.
			if( controller.getCurrentAgent() != null && controller.getCurrentAgent() instanceof User ){
				if( x_offset <= 37 && x_offset >= -5 && y_offset <= 37 && y_offset >= -5){


					// Check time to see if it is double-click. double-click if < .75 sec.
					if( ((new Date()).getTime() - doubleclick.getTime()) < 750 ){

						// Close old PropDialog if it exists.
						try{
							propDialog.dispose();
						}catch(Exception exc){}

						// Open a PropDialog
						propDialog = new PropDialog( controller, controller.getFrame() );
						controller.fireControllerChange();

					}else{
						//Get time of click for checking for double-click next time.
						doubleclick = new Date();
					}
				}

			} //if

		} //if


	} //method: mouseClicked





    /* ===================================================================== */
    /** This method is used to get the offset position when we click and drag
     *  an agent on the map.  Offset is needed for correctly drawing the agent
     *  on the map.
     *
     *  @param  e   The incoming MouseEvent containing button and point info.
     *
     */
	public void mousePressed(MouseEvent e){


		int btn = e.getButton();
		Point p = e.getPoint();

		// See if we are clicking and dragging with left button.
		if( btn == 1 )
			clicked = true;

		// Calulate the offset from agent location and the click point.
		if( controller.getCurrentAgent() != null ){
			Point agent = controller.getCurrentAgent().getLocation();

			x_offset = (int)(p.getX() - agent.getX());
			y_offset = (int)(p.getY() - agent.getY());

		}


	} //method: mousePressed




    /* ===================================================================== */
    /** This method lets the object know that the user has released from a
     *  mouse drag.  It sends a location change event, so that the location
     *  of the current user can be sent to the iPAQ application.
     *
     *  @param  e   The incoming MouseEvent containing button and point info.
     *
     */
	public void mouseReleased(MouseEvent e){

		int btn = e.getButton();
		Point p = e.getPoint();


		if( btn == 1 )
			clicked = false;

		// Make sure there is an agent, set new GPS location, and send
		// the change notification.
		if( controller.getCurrentAgent() != null ){

			Position old = controller.getCurrentAgent().getPosition();
			Position chart_pos = chartdata.transPos(p);

			chart_pos.setHeading( old.getHeading() );
			chart_pos.setTime( old.getTime() );


			controller.getCurrentAgent().setPosition( chart_pos );

			if( controller.getCurrentAgent() instanceof User ){
				controller.locationChange(((User)(controller.getCurrentAgent())));
			}

			this.setCursor( Cursor.getDefaultCursor() );


			controller.fireControllerChange();

		} //if


	} //method: mouseReleased




    /* ===================================================================== */
    /** This method is for moving an Agent around on the map.  It determines
     *  the closest Agent to the drag point, determines if that point is on
     *  an Agent, and then redraws the Agent to the new location based on
     *  offsets gotten from mousePressed method.
     *
     *  @param  e   The incoming MouseEvent containing button and point info.
     *
     */
	public void mouseDragged(MouseEvent e){

		Point p = e.getPoint();
		int btn = e.getButton();


		if( btn == 0 ){

			//System.out.println("point: "+ p.getX() + "    "+p.getY() );

			double d = 100000;
			double temp;

			// Find the closest User or AccessPoint.
			for(int i=0; i<controller.getUsers().size(); i++){

				temp = diff(p, ((User)controller.getUsers().elementAt(i)).getLocation());

				if( temp < d ){
					d = temp;
					controller.setCurrentAgent( ((User)controller.getUsers().elementAt(i)) );
				}

			}
			for(int i=0; i<controller.getPoints().size(); i++){

				temp = diff(p, ((Agent)controller.getPoints().elementAt(i)).getLocation());

				if( temp < d ){
					d = temp;
					controller.setCurrentAgent( ((Agent)controller.getPoints().elementAt(i)) );
				}

			}

			// Make sure the closest Agent is VERY close to the drag point.
			if( controller.getCurrentAgent() != null ){
				if( x_offset <= 37 && x_offset >= -5 && y_offset <= 37 && y_offset >= -5){

					// Do not allow dragging to off the map.
					if( (int)p.getX()-x_offset > -15 && (int)p.getY()-y_offset > -15
					   && (int)p.getX()-x_offset < (this.getSize().getWidth()-20)
					   && (int)p.getY()-y_offset < (this.getSize().getHeight()-20) ){

						// Get new location and repaint.
						p = new Point((int)p.getX()-x_offset, (int)p.getY()-y_offset);
						controller.getCurrentAgent().setLocation(p);

						repaint();
					}
				}

			} //if

		} //if( btn == 0 )

	} //method: mouseDragged


    /* ===================================================================== */
    /** This method not used.
     *
     *  @param  e
     *
     */
 	public void mouseMoved(MouseEvent e){} //method: mouseMoved
    /* ===================================================================== */
    /** This method not used.
     *
     *  @param  e
     *
     */
	public void mouseEntered(MouseEvent e){} //method: mouseEntered
    /* ===================================================================== */
    /** This method not used.
     *
     *  @param  e
     *
     */
	public void mouseExited(MouseEvent e){} //method: mouseExited




    /* ===================================================================== */
    /** This method to tell this object whether or not it should paint the
     *  wireless range for each agent in the simulation.
     *
     *  @param  ranges   True if want to show wireless ranges.
     *
     */
	public void setRanges(boolean ranges){

		this.ranges = ranges;

		// Show the wireless ranges now.
		repaint();

	} //method: setRanges



    /* ===================================================================== */
    /** This method returns the pixel distance between two points.
     *
     *  @param  p1   The first point.
     *  @param  p2   The second point.
     *
     */
	private double diff(Point p1, Point p2){

		double x = p1.getX() - p2.getX();
		if( x < 0 )
			x = x * -1;

		double y = p1.getY() - p2.getY();
		if( y < 0 )
			y = y * -1;


		return Math.sqrt( x*x + y*y );

	} //method: diff




    /* ===================================================================== */
    /** This method returns the chartdata kept inside this object used for
     *  turning pixel coordinates into real GPS coordinates.
     *
     *  @return  ChartData   The chart data used in this object to return.
     *
     */
	public ChartData getChartData(){

		return chartdata;

	} //method: getChartData




    /* ===================================================================== */
    /** This method loads new chart data, by first reading the given xml file
     *  which contains map name and reference point information.  Then the new
     *  ChartData will give us the map image to use.
     *
     *  @param  name   file name for new data to use.
     *
     */
	public void loadChart(String name){

		this.setVisible(false);

		this.readData(name);

		if (img!=null) {img.flush();}

		img = getToolkit().getImage("files/maps/"+chartdata.imgFileName);


		this.setSize(this.getPreferredSize());


		this.setVisible(true);
		repaint();

	} //method: loadChart



    /* ===================================================================== */
    /** This method loads new chart data, by first reading the given xml file
     *  which contains map name and reference point information.
     *
     *  @param  file_name   Name of XML file that contains new ChartData info.
     *
     */
 	public void readData(String file_name){

    	chartdata = null;

		InputParser ip = new InputParser(file_name);

		try{
			chartdata=ChartData.loadChartData(ip);
		}catch (IOException e) {}


  	} //method: readData


  /**
   * Add the listener to the listeners list 
   */
  public void addChartCompListener(ChartCompListener listener) {
    if (!listeners.contains(listener))
      listeners.add(listener);
  }

  /**
   * Fires the GraphicsPaintedEvent to all the listeners
   */
  private void fireGraphicsPaintedEvent(Graphics g) {
    GraphicsPaintedEvent e = new GraphicsPaintedEvent(g);
    for (int i = 0; i < listeners.size(); i++)
      ((ChartCompListener)listeners.elementAt(i)).graphicsPainted(e);
  }

} //class: ChartComp

