// PnlTabRoute.java



package gui.tray;



import gui.*;

import agent.*;

import util.*;

import controller.*;

import transport.*;

import transport.protocol.*;



import java.awt.*;

import java.awt.event.*;

import javax.swing.*;

import javax.swing.border.*;

import javax.swing.table.*;

import javax.swing.event.*;

import java.util.*;

import java.io.*;



/**

 * This tab allows us to assign routes to the users and optionally see them

 * on the map. It also allows us to create new routes. 

 *

 * @author Amit G.

 */

public class PnlTabRoute extends JPanel implements 

  MouseListener, ActionListener, ChartCompListener, ControllerListener {



  /** The simulation controller that keeps all data. */

  private Controller controller;



  private JButton createBtn; 



  /** Current route being plotted */

  private Route currentRoute;



  /** Name of the route being plotted */

  private String routeName = new String();



  /** Indicates if the plot mode on */

  private boolean isPlotMode = false;



  /** The route table which displays the route for each user */

  private JTable          table;

  private RouteTableModel tableModel;



  /** The color in which the routes are plotted */

  private final Color LINE_COLOR = new Color(255,128,0);

  private final int   LINE_WIDTH = 2;



  /** 

   * User Route colors - the routes for every user are drawn using the

   * distinctive color for that user - which is the color of their tie!

   */

  public static final Color[] ROUTE_COLORS = { Color.red, 

                                               Color.green, 

                                               Color.yellow,

                                               new Color(155,79,163), 

                                               new Color(63,187,239), 

                                               Color.black, 

                                               new Color(255,237,236),

                                               new Color(190,222,51)

                                             };



  /** Creates the "Route" tab */

  public PnlTabRoute(Controller controller) {

    this.controller = controller;



    controller.addControllerListener(this);



    // Add to ChartComp as a mouseListener and ChartCompListener

    controller.getChartComp().addMouseListener(this);

    controller.getChartComp().addChartCompListener(this);



    setLayout(new BorderLayout());



    // create and add the button

    createBtn = new JButton("Create new route");

    createBtn.addActionListener(this);

    add(createBtn, BorderLayout.SOUTH);



    // create the table

    tableModel = new RouteTableModel();

    tableModel.addTableModelListener(new TableModelListener() {

        public void tableChanged(TableModelEvent e) {

          PnlTabRoute.this.controller.getChartComp().repaint();

        }

      });

    table = new JTable(tableModel);

    table.setRowHeight(20);



    // a combo box is used for route selection

    JComboBox combo = new JComboBox();

    table.getColumnModel().getColumn(1).setCellEditor(new ComboEditor(combo));

    

    JScrollPane sp = new JScrollPane(table, 

                                     JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,

                                     JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

    sp.setBorder(BorderFactory.createTitledBorder("User Routes"));



    add(sp, BorderLayout.CENTER);

  }



  /**

   * Listens for button-click on the "Create new route" button. If plotting

   * is started, starts listening for mouse clicks on the map. If plotting

   * is stopped, prompts for file name to store the route.

   */

  public void actionPerformed(ActionEvent ev) {



    // start the plotting mode

    if (!isPlotMode) {

      isPlotMode = true;

      table.setEnabled(false);

      createBtn.setText("Save route");

      currentRoute = new Route(controller.getChartComp().getChartData()

                        .getImgFileName());

    }

    

    // stop mapping mode, write the route to a file

    else { 

      isPlotMode = false;

      table.setEnabled(true);

      createBtn.setText("Create new route");



      // proceed only if at least a line was plotted

      if (currentRoute.getPoints().size() > 1) {

        routeName = JOptionPane.showInputDialog(PnlTabRoute.this,

                                                "Save route as",

                                                routeName);

        try {

          if (routeName != null) currentRoute.write(routeName);

        }

        catch (Exception e) {

          System.err.println(e);

          //e.printStackTrace();//turn this on for line number

        }

      }



      // clear the route from the map 

      controller.getChartComp().repaint();

      currentRoute = null;

    }

  }





  /**

   * Listens for mouse-clicks on the map (only if the route is being mapped)

   * It then draws the route plotted so far. A left-button click draws the

   * route while any other button click removes the last drawn point

   */

  public void mouseClicked(MouseEvent e) {

    if (!isPlotMode) return;



    if (e.getButton() == MouseEvent.BUTTON1) {

      // record the point and position

      Point    p    = e.getPoint();

      Position posn = controller.getChartComp().getChartData().transPos(p);

      currentRoute.addPoint(p, posn);

    }

    else {

      currentRoute.removeLastPoint();

    }

    controller.getChartComp().repaint();

  }

  public void mouseEntered(MouseEvent e)  {}

  public void mouseExited(MouseEvent e)   {}

  public void mousePressed(MouseEvent e)  {}

  public void mouseReleased(MouseEvent e) {}



  /**

   * The event handler for when the graphics on the map are painted. 

   * For e.g when the map is scrolled, window restored etc. This method

   * redraws all the routes.

   */

  public void graphicsPainted(GraphicsPaintedEvent e) {



    // draw the routes for all the users

    for (int i = 0; i < tableModel.getRowCount(); i++) {

      RouteUser ru = tableModel.getRouteUserAt(i);

      if (ru.isVisible())

        ru.draw((Graphics)e.getSource());

    }

    

    // if a route was being plotted, draw it

    if (isPlotMode)

      currentRoute.draw((Graphics)e.getSource(),LINE_COLOR,LINE_WIDTH);

  }



  /**

   * Method called by controller when user list is updated

   */

  public void refresh() { 

    tableModel.setUsers(controller.getUsers());

  }





  // binds together route and User (agent)

  class RouteUser{

    

    private User user;

    private String route;

    private Boolean visible;

    

    public RouteUser(User user, String route, Boolean visible) {

      this.user = user;

      this.route = route;

      this.visible = visible;

    }



    public RouteUser(User user, String route) {

      this(user, route, Boolean.FALSE);

    }



    public RouteUser(User user) {

      this(user, "");

    }



    public User    getUser()    { return user; }

    public String  getRoute()   { return route; }

    public Boolean getVisible() { return visible; }

    public boolean isVisible()  { return visible.booleanValue(); }



    public void setRoute(String route) { 

      this.route = route;

      

      // send message to the User

      RouteChangeMessage rm = new RouteChangeMessage(user.getPort(), 2323,

                                                     route);

      SocketSender ss = new SocketSender(rm);

      ss.start();

    }



    public void setVisible(Boolean visible) { this.visible = visible; }



    /** Draws this route using the graphics object passed in */

    public void draw(Graphics g) {

      if (route.equals("")) return;



      // get the route color for this user based on his route name

      String uname = user.getName();

      int colorIndex = Integer.parseInt(uname.substring(uname.length() - 1));



      // now draw the route

      Route r = new Route(new File(route));

      r.draw(g, ROUTE_COLORS[colorIndex], LINE_WIDTH);

    }

  } // class RouteUser



  /**

   * Provides the data for the table.

   */

  class RouteTableModel extends AbstractTableModel {



    private Vector routeUsers = new Vector();



    public String getColumnName(int columnIndex) {

      switch (columnIndex) {

      case 0: return "User";

      case 1: return "Route";

      case 2: return "Show";

      }

      return null;

    }

    

    public Object getValueAt(int row, int col) {

      RouteUser ur = (RouteUser)routeUsers.elementAt(row);

      switch (col) {

      case 0: return ur.getUser().getName();

      case 1: return ur.getRoute();

      case 2: return ur.getVisible();

      }

      return null;

    }



    public void setValueAt(Object aValue, int rowIndex, int colIndex) {

      RouteUser ur = (RouteUser)routeUsers.elementAt(rowIndex);

      switch (colIndex) {

      case 1: ur.setRoute((String)aValue);    break;

      case 2: ur.setVisible((Boolean)aValue); break;

      }

      fireTableCellUpdated(rowIndex, colIndex);

    }

    

    public int getColumnCount() { return 3; }

    public int getRowCount()    { return routeUsers.size();    }



    public boolean isCellEditable(int rowIndex, int colIndex) {

      return (colIndex == 0)? false : true;

    }



    public Class getColumnClass(int colIndex) {

      switch(colIndex) {

      case 0:

      case 1: return String.class;

      case 2: return Boolean.class;

      }

      return null;

    }



    public void setUsers(Vector u) {



      Vector newUsers = new Vector();

      

      // take every user in the vector passed in and check if he 

      // already exists in our list. If not, create a RouteUser object

      // for that user and add it to our new list. If he does already exist,

      // add him to the new list as well. This ensures that the new users are

      // added and the users deleted are left out. All this is mainly to

      // maintain the route names of the users

      for (int i = 0; i < u.size(); i++) {

        User userObj = (User)u.elementAt(i);

        boolean newuser = true;

        for (int j = 0; j < routeUsers.size(); j++) {

          RouteUser oldone = (RouteUser)routeUsers.elementAt(j);

          if (oldone.getUser().getName().equals(userObj.getName())) {

            newUsers.add(oldone);

            newuser = false;

            break;

          }

        }

        if (newuser) 

          newUsers.add(new RouteUser(userObj));

      }



      routeUsers = newUsers;

      fireTableDataChanged();

    }



    public RouteUser getRouteUserAt(int index) {

      return (RouteUser)routeUsers.elementAt(index);

    }



  } // class RouteTableModel





  /**

   * The combo box used to select a route. The reason why DefaultCellEditor

   * is not directly used is because the combo box model might change during

   * execution when new routes are added

   */

  class ComboEditor extends DefaultCellEditor {

    

    ComboEditor(JComboBox combo) {

      super(combo);

    }



    public Component getTableCellEditorComponent(JTable table,

                                                 Object value,

                                                 boolean isSelected,

                                                 int row,

                                                 int column)

    {

      JComboBox c = (JComboBox)super.getTableCellEditorComponent(table,

                                                                 value,

                                                                 isSelected,

                                                                 row,

                                                                 column);

      c.setModel(new DefaultComboBoxModel(getRoutes()));

      c.setSelectedItem(value);

      return c;

    }

    

    // scans the routes dir and returns all the valid route names

    private String[] getRoutes() {

      File dir = new File(Route.ROUTES_DIR);

      String[] files = dir.list(new FilenameFilter() {

          public boolean accept(File dir, String name) {

            return name.endsWith(".xml");

          }

        });

      String[] finalList = new String[files.length + 1];

      finalList[0] = ""; // add one element with empty string

      System.arraycopy(files, 0, finalList, 1, files.length);

      return finalList;

    }

    

  }



} // class PnlTabRoute

