// Route.java

package util;

import java.util.*;
import java.io.*;
import java.awt.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

/**
 * Encapsulates a route on the map. Stores the map filename, pixel positions
 * and absolute (geograhical) positions. The route is stored in xml format in
 * the directory given by ROUTES_DIR. The route name is always the xml file
 * in which it is stored. 
 *
 * @author Amit G.
 */
public class Route {

  /** Vector of Point objects. Each point denotes a pixel position on the map*/
  private Vector points = new Vector();

  /** Vector of Position objects. Each point denotes a geographical position */
  private Vector positions = new Vector();

  /** Name of the route (Also the filename where its stored)*/
  private String routeName;

  /** File name of the map */
  private String mapName;

  /** Default directory for the routes */
  public static final String ROUTES_DIR = "files/routes/";

  /** 
   * Create a new Route using the name and map file name passed in. If the 
   * route already exists, it will be overwritten on the next call to 
   * write()
   */
  public Route(String routeName, String mapName) {
    this.mapName = mapName;
    setRouteName(routeName);
  }

  /**
   * Creates a new Route without any points.
   * @params mapName The map associated with this Route
   */
  public Route(String mapName) {
    this.mapName = mapName;
  }

  /**
   * Reads the <code>routeFile</code> and creates a Route based on data
   * in that file.
   * @params routeFile The xml file containing the route data
   */
  public Route(File routeFile) {
    routeName = routeFile.getName();
    
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    factory.setNamespaceAware(false);
    factory.setValidating(false);
    
    DocumentBuilder builder;
    Document document = null;
    
    try {
      builder  = factory.newDocumentBuilder();
      document = builder.parse(ROUTES_DIR + routeFile);
      
      Element root = (Element)document.getDocumentElement();

      Element mapE = (Element)root.getElementsByTagName("mapname").item(0);
      mapName = ((Text)mapE.getChildNodes().item(0)).getNodeValue();

      NodeList list= root.getElementsByTagName("point");
      for (int i = 0; i < list.getLength(); i++) {
        Element n = (Element)list.item(i);
        points.add(i, new Point((int)Float.parseFloat(n.getAttribute("x")), 
                                (int)Float.parseFloat(n.getAttribute("y"))));

        Position pos = new Position();
        pos.setLatitude(Double.parseDouble(n.getAttribute("latitude")));
        pos.setLongitude(Double.parseDouble(n.getAttribute("longitude")));
        positions.add(pos);
      }
    }
    catch(Exception e) {
      System.err.println("Route.<init>(File): " + e);
      e.printStackTrace();//turn on for line number
    }
    
  }

  /**
   * Adds <code>point</code> and <code>position</code> to the end of this
   * route. 
   */
  public void addPoint(Point point, Position position) {
    points.add(point);
    positions.add(position);
  }

  /**
   * Writes this route in XML format to the file <pre>fileName</pre>
   * overwriting existing file. It adds the .xml extension to the filename 
   * if not already present
   *
   * @param fileName The name of the file where the route is written
   */
  public void write(String fileName) throws Exception { 

    // Create the document
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    factory.setNamespaceAware(false);
    factory.setValidating(false);
    factory.setCoalescing(false);
    factory.setIgnoringElementContentWhitespace(true);
    
    DocumentBuilder builder = factory.newDocumentBuilder();
    Document document = builder.newDocument();
    
    // add the preliminary nodes
    Element root = document.createElement("route");
    document.appendChild(root);
    
    Element mapNameElem = document.createElement("mapname");
    mapNameElem.appendChild(document.createTextNode(mapName));
    root.appendChild(mapNameElem);

    Element pointsRoot = document.createElement("points");
    root.appendChild(pointsRoot);

    // add the points and positions to the document
    ListIterator pointsI    = points.listIterator();
    ListIterator positionsI = positions.listIterator();
    while (pointsI.hasNext()) {
      Point    p   = (Point)pointsI.next();
      Position pos = (Position)positionsI.next();

      Element pointE = document.createElement("point");
      pointE.setAttribute("x"        , String.valueOf(p.getX()));
      pointE.setAttribute("y"        , String.valueOf(p.getY()));
      pointE.setAttribute("latitude" , String.valueOf(pos.getLatitude()));
      pointE.setAttribute("longitude", String.valueOf(pos.getLongitude()));
      pointE.setAttribute("time"     , ""); //blank for now

      pointsRoot.appendChild(pointE);
    }

    // start writing the file
    if (!fileName.endsWith(".xml")) fileName += ".xml";
    fileName = ROUTES_DIR + fileName;

    System.out.println("writing " + fileName);
    Transformer  trans  = TransformerFactory.newInstance().newTransformer();
    DOMSource    dsrc   = new DOMSource(document);
    StreamResult result = new StreamResult(new FileOutputStream(fileName));
    trans.transform(dsrc, result);
    result.getOutputStream().close();
  }

  /**
   * Saves this route to the default route file overwriting previous contents.
   */ 
  public void write() throws Exception {
    if (routeName == null) 
      throw new Exception("Route name is not set");
    write(routeName);
  }

  /**
   * Draws the current route on the Graphics object passed in.
   */
  public void draw(Graphics graphics, Color color, int lineWidth) {
    Graphics2D g = (Graphics2D)graphics;
    g.setColor(color);

    Point lp = null; // last drawn point

    for (int i = 0; i < points.size(); i++) {
      Point p = (Point)points.elementAt(i);
      g.fillOval((int)p.getX() - 5, (int)p.getY() - 5, 10, 10);
      if (lp != null) {
        g.setStroke(new BasicStroke(lineWidth));
        g.drawLine((int)lp.getX(),
                   (int)lp.getY(), 
                   (int)p.getX(),
                   (int)p.getY());
      }
      lp = p;
    }
  }

  /**
   * Deletes the last point from the route
   */
  public void removeLastPoint() { 
    if (points.size() == 0) return;
    points.removeElementAt(points.size() - 1);
    positions.removeElementAt(positions.size() - 1);
  }

  /**
   * Sets the route name to <code>routeName</code>
   */
  public void setRouteName(String routeName) {
    this.routeName = routeName;
    if (!this.routeName.endsWith(".xml")) 
      this.routeName += ".xml";
  }
  

  public String getName()      { return routeName; }
  public String getMapName()   { return mapName; }
  public Vector getPoints()    { return points; }
  public Vector getPositions() { return positions; }

}
